# -*- coding: utf-8 -*-
from .attr_symbols import AttributeSymbolChain, CallPoint, get_attribute_symbol_chain
from .lineno_stmt_map import compute_lineno_to_stmt_mapping
from .precheck import *
from .stmt_edges import *



# -*- coding: utf-8 -*-
import os
from typing import TYPE_CHECKING

from IPython import get_ipython
import pytest

from nbsafety.safety import DependencySafety

if TYPE_CHECKING:
    from typing import Any, List, Optional, Tuple


def should_skip_known_failing(reason='this test tests unimpled functionality'):
    return {
        'condition': os.environ.get('SHOULD_SKIP_KNOWN_FAILING', True),
        'reason': reason
    }


skipif_known_failing = pytest.mark.skipif(**should_skip_known_failing())


def assert_bool(val, msg=''):
    assert val, str(msg)


# Reset dependency graph before each test to prevent unexpected stale dependency
def make_safety_fixture(**kwargs) -> 'Tuple[Any, List[Optional[DependencySafety]], Any]':
    safety_state: List[Optional[DependencySafety]] = [None]

    def run_cell(code):
        get_ipython().run_cell_magic(safety_state[0].cell_magic_name, None, code)

    @pytest.fixture(autouse=True)
    def init_or_reset_dependency_graph():
        safety_state[0] = DependencySafety(**kwargs)
        run_cell('import logging')
        yield  # yield to execution of the actual test
        get_ipython().reset()  # reset ipython state

    return init_or_reset_dependency_graph, safety_state, run_cell



# -*- coding: utf-8 -*-
from ipykernel.ipkernel import IPythonKernel
import nbsafety.safety
from ..safety import DependencySafety
from ..version import __version__
# import warnings
#Avoid networkx gives a FutureWarning
# TODO (smacke): this disables all FutureWarnings; disable until we can figure out a better system
# warnings.filterwarnings("ignore", category=FutureWarning)

_SAFETY_STATE = '__SAFETY_STATE'
_CELL_MAGIC_NAME = '__SAFETY_CELL_MAGIC'
_LINE_MAGIC_NAME = '__SAFETY_LINE_MAGIC'


class SafeKernel(IPythonKernel):
    implementation = 'kernel'
    implementation_version = __version__

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_safety()

    def _init_safety(self):
        self.shell.run_cell(f'from {nbsafety.safety.__name__} import {DependencySafety.__name__}')
        self.shell.run_cell(
            f'{_SAFETY_STATE} = {DependencySafety.__name__}'
            f'(cell_magic_name="{_CELL_MAGIC_NAME}",'
            f' line_magic_name="{_LINE_MAGIC_NAME}",'
            f' store_history=False'
            f')'
        )
        # hack to init logging
        # TODO: figure out why this is necessary
        self.do_execute("""
def _foo():
    pass
_foo()  # for some reason the fn call is necessary to properly init logging
del _foo
""", False)
        self.shell.run_cell(f'{_SAFETY_STATE}._logging_inited()')

    def do_execute(self, code, silent, store_history=False,
                   user_expressions=None, allow_stdin=False):
        if code.split()[0] == "%safety":
            code = "%{} {}".format(_LINE_MAGIC_NAME, ' '.join(code.split()[1:]))
            return super().do_execute(code, silent, False, user_expressions, allow_stdin)
        code = f"%%{_CELL_MAGIC_NAME}\n{code}"
        return super().do_execute(code, silent, False, user_expressions, allow_stdin)



# -*- coding: utf-8 -*-
import json
import os
import sys
import argparse

from jupyter_client.kernelspec import KernelSpecManager
from IPython.utils.tempdir import TemporaryDirectory

PACKAGE = __package__.split('.')[0]
DISPLAY_NAME = f"Python 3 ({PACKAGE})"
kernel_json = {
 "argv": [
   sys.executable, "-m", "nbsafety.kernel", "-f", "{connection_file}",
 ],
 "display_name": DISPLAY_NAME,
 "language": "python",
 "codemirror_mode": "shell",
}


def install_my_kernel_spec(user=True, prefix=None):
    with TemporaryDirectory() as td:
        os.chmod(td, 0o755)  # Starts off as 700, not user readable
        with open(os.path.join(td, 'kernel.json'), 'w') as f:
            json.dump(kernel_json, f, sort_keys=True)
        # TODO: Copy resources once they're specified

        print(f'Installing KernelSpec for {PACKAGE} kernel')
        KernelSpecManager().install_kernel_spec(td, f'python3-{PACKAGE}', user=user, prefix=prefix)


def _is_root():
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False  # assume not an admin on non-Unix platforms


def main(argv=None):
    parser = argparse.ArgumentParser(
        description=f'Install KernelSpec for {PACKAGE} kernel.'
    )
    prefix_locations = parser.add_mutually_exclusive_group()

    prefix_locations.add_argument(
        '--user',
        help='Install KernelSpec in user homedirectory',
        action='store_true'
    )
    prefix_locations.add_argument(
        '--sys-prefix',
        help='Install KernelSpec in sys.prefix. Useful in conda / virtualenv',
        action='store_true',
        dest='sys_prefix'
    )
    prefix_locations.add_argument(
        '--prefix',
        help='Install KernelSpec in this prefix',
        default=None
    )

    args = parser.parse_args(argv)

    user = False
    prefix = None
    if args.sys_prefix:
        prefix = sys.prefix
    elif args.prefix:
        prefix = args.prefix
    elif args.user or not _is_root():
        user = True

    install_my_kernel_spec(user=user, prefix=prefix)
    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
import glob
import importlib
import os


pkg = os.path.basename(os.path.dirname(__file__))
for mod in glob.glob(os.path.join(os.path.dirname(__file__), "test_*.py")):
    mod = os.path.basename(mod)[:-3]
    importlib.import_module(pkg + '.' + mod, '.')



# -*- coding: utf-8 -*-
import ast
from contextlib import contextmanager
import logging
import sys
from typing import TYPE_CHECKING

from IPython.core.magic import register_cell_magic, register_line_magic

from .analysis import AttributeSymbolChain, precheck, compute_lineno_to_stmt_mapping
from .ipython_utils import (
    ast_transformer_context,
    cell_counter,
    run_cell,
    save_number_of_currently_executing_cell,
)
from . import line_magics
from .scope import Scope
from .tracing import AttributeTracingManager, make_tracer, TraceState

if TYPE_CHECKING:
    from typing import Dict, Set, Optional
    from .data_cell import DataCell

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def _safety_warning(name: str, defined_cell_num: int, required_cell_num: int, fresher_ancestors: 'Set[DataCell]'):
    logger.warning(
        f'{name} defined in cell {defined_cell_num} may depend on '
        f'old version(s) of [{", ".join(str(dep) for dep in fresher_ancestors)}] '
        f'(lastest update in cell {required_cell_num}).'
    )


class DependencySafety(object):
    """Holds all the state necessary to detect stale dependencies in Jupyter notebooks."""
    def __init__(self, cell_magic_name=None, line_magic_name=None, **kwargs):
        self.global_scope = Scope()
        self.namespaces: Dict[int, Scope] = {}
        self.statement_cache: Dict[int, Dict[int, ast.stmt]] = {}
        self.stale_dependency_detected = False
        self.trace_state = TraceState(self)
        self.attr_trace_manager = AttributeTracingManager(self.namespaces, self.global_scope)
        self.store_history = kwargs.pop('store_history', True)
        self.trace_messages_enabled = kwargs.pop('trace_messages_enabled', False)
        self._save_prev_trace_state_for_tests = kwargs.pop('save_prev_trace_state_for_tests', False)
        if self._save_prev_trace_state_for_tests:
            self.prev_trace_state: Optional[TraceState] = None
        self._cell_magic = self._make_cell_magic(cell_magic_name)
        # Maybe switch update this too when you are implementing the usage of cell_magic_name?
        self._line_magic = self._make_line_magic(line_magic_name)
        self._last_refused_code: Optional[str] = None
        self._last_cell_ast: Optional[ast.Module] = None
        self._track_dependencies = True

        self._disable_level = 0

    def _logging_inited(self):
        self.store_history = True
        logger.setLevel(logging.WARNING)

    def _precheck_for_stale(self, cell):
        # Precheck process. First obtain the names that need to be checked. Then we check if their
        # `defined_cell_num` is greater than or equal to required; if not we give a warning and return `True`.
        self._last_cell_ast = ast.parse('\n'.join(
            [line for line in cell.strip().split('\n') if not line.startswith('%') and not line.endswith('?')])
        )
        self.statement_cache[cell_counter()] = compute_lineno_to_stmt_mapping(self._last_cell_ast)
        if self._last_refused_code is None or cell != self._last_refused_code:
            for name in precheck(self._last_cell_ast, self.global_scope.data_cell_by_name.keys()):
                if isinstance(name, str):
                    nodes = [self.global_scope.data_cell_by_name.get(name, None)]
                elif isinstance(name, AttributeSymbolChain):
                    nodes = self.global_scope.gen_data_cells_for_attr_symbol_chain(name, self.namespaces)
                else:
                    logger.warning('invalid type for name %s', name)
                    continue
                for node in nodes:
                    if node is None:
                        continue
                    if node.is_stale() and self._disable_level < 2:
                        _safety_warning(name, node.defined_cell_num, node.required_cell_num, node.fresher_ancestors)
                        self.stale_dependency_detected = True
                        self._last_refused_code = cell
                        if self._disable_level == 0:
                            return True
        else:
            # TODO: break dependency chain here
            pass

        self._last_refused_code = None
        return False

    def _make_cell_magic(self, cell_magic_name):
        def _dependency_safety(_, cell: str):
            if self._disable_level == 3:
                run_cell(cell)
                return

            with save_number_of_currently_executing_cell():
                # Stage 1: Precheck.
                if self._precheck_for_stale(cell):
                    return

                def _backup():
                    # something went wrong silently (e.g. due to line magic); fall back to just executing the code
                    logger.warning('Something failed while attempting traced execution; '
                                   'falling back to uninstrumented execution.')
                    run_cell(cell, store_history=self.store_history)

                # Stage 2: Trace / run the cell, updating dependencies as they are encountered.
                with self._tracing_context(untraced_backup=_backup):
                    run_cell(cell, store_history=self.store_history)

        if cell_magic_name is not None:
            # TODO (smacke): probably not a great idea to rely on this
            _dependency_safety.__name__ = cell_magic_name
        return register_cell_magic(_dependency_safety)

    @contextmanager
    def _tracing_context(self, untraced_backup=None):
        sys.settrace(make_tracer(self))
        try:
            with ast_transformer_context(self.attr_trace_manager.ast_transformer):
                yield
        finally:
            sys.settrace(None)
            # TODO: add more explicit way to check for an error in dependency tracing code
            if self.trace_state.prev_trace_stmt_in_cur_frame is None:
                if untraced_backup is not None:
                    untraced_backup()
            else:
                self._reset_trace_state_hook()

    def _reset_trace_state_hook(self):
        if self.dependency_tracking_enabled:
            self.trace_state.prev_trace_stmt_in_cur_frame.finished_execution_hook()
        self.attr_trace_manager.reset()
        if self._save_prev_trace_state_for_tests:
            self.prev_trace_state = self.trace_state
        self.trace_state = TraceState(self)

    def _make_line_magic(self, line_magic_name):
        def _safety(line_: str):
            line = line_.split()
            if not line or line[0] not in [
                "show_graph", "show_dependency", "show_stale", "set_disable_level", "trace_messages",
                "remove_dependency", "add_dependency", "turn_off_warnings_for", "turn_on_warnings_for",
            ]:
                print(line_magics.USAGE)
                return
            if line[0] == "show_graph":
                return line_magics.show_graph(self)
            elif line[0] == "show_dependency":
                return line_magics.show_deps(self, line)
            elif line[0] == "show_stale":
                return line_magics.show_stale(self)
            elif line[0] == "set_disable_level":
                return line_magics.set_disable_level(self, line)
            elif line[0] == "trace_messages":
                return line_magics.configure_trace_messages(self, line)
            elif line[0] == "remove_dependency":
                return line_magics.remove_dep(self, line)
            elif line[0] == "add_dependency":
                return line_magics.add_dep(self, line)
            elif line[0] == "turn_off_warnings_for":
                return line_magics.turn_off_warnings_for(self, line)
            elif line[0] == "turn_on_warnings_for":
                return line_magics.turn_on_warnings_for(self, line)

        if line_magic_name is not None:
            _safety.__name__ = line_magic_name
        return register_line_magic(_safety)

    @property
    def dependency_tracking_enabled(self):
        return self._track_dependencies

    @property
    def cell_magic_name(self):
        return self._cell_magic.__name__

    @property
    def line_magic_name(self):
        return self._line_magic.__name__

    def test_and_clear_detected_flag(self):
        ret = self.stale_dependency_detected
        self.stale_dependency_detected = False
        return ret



# -*- coding: utf-8 -*-
import ast
from typing import TYPE_CHECKING

# from .attr_symbols import get_attribute_symbol_chain
from .mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin

if TYPE_CHECKING:
    from typing import List, Set
    from .attr_symbols import AttributeSymbolChain


class GetStatementLvalRvalSymbols(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        # TODO: current complete bipartite subgraph will add unncessary edges
        self.lval_symbol_set: Set[str] = set()
        self.rval_symbol_set: Set[str] = set()
        self.lval_attr_chains: List[AttributeSymbolChain] = []
        self.rval_attr_chains: List[AttributeSymbolChain] = []
        self.gather_rvals = True

    def __call__(self, node):
        self.visit(node)
        return self.lval_symbol_set, self.rval_symbol_set

    @property
    def to_add_set(self):
        if self.gather_rvals:
            return self.rval_symbol_set
        else:
            return self.lval_symbol_set

    @property
    def to_append_list(self):
        if self.gather_rvals:
            return self.rval_attr_chains
        else:
            return self.lval_attr_chains

    def gather_lvals_context(self):
        return self.push_attributes(gather_rvals=False)

    def gather_rvals_context(self):
        return self.push_attributes(gather_rvals=True)

    def visit_Attribute(self, node):
        # skip node.attr -- this is handled by the attribute tracer
        # also only add rvals -- lvals will also be handled by tracer
        with self.gather_rvals_context():
            self.visit(node.value)

    def visit_Name(self, node):
        self.to_add_set.add(node.id)

    def visit_Subscript(self, node: ast.Subscript):
        self.visit(node.value)
        with self.gather_rvals_context():
            self.visit(node.slice)

    def visit_Assign(self, node):
        with self.gather_lvals_context():
            for target in node.targets:
                self.visit(target)
        self.visit(node.value)

    def visit_AugAssign(self, node):
        with self.gather_lvals_context():
            self.visit(node.target)
        with self.gather_rvals_context():
            self.visit(node.value)

    def visit_For(self, node):
        # skip body -- will have dummy since this visitor works line-by-line
        with self.gather_lvals_context():
            self.visit(node.target)
        with self.gather_rvals_context():
            self.visit(node.iter)

    def visit_FunctionDef(self, node):
        self.lval_symbol_set.add(node.name)
        with self.gather_rvals_context():
            self.visit(node.args)

    def visit_ClassDef(self, node):
        self.lval_symbol_set.add(node.name)
        with self.gather_rvals_context():
            self.visit(node.bases)
            self.visit(node.decorator_list)

    def visit_Keyword(self, node):
        self.visit(node.value)

    def visit_Starred(self, node):
        self.visit(node.value)

    def visit_AsyncFunctionDef(self, node):
        self.visit_FunctionDef(node)

    def visit_Lambda(self, node):
        assert self.gather_rvals
        # remove node.arguments
        self.visit(node.body)
        self.visit(node.args)
        with self.push_attributes(rval_symbol_set=set()):
            self.visit(node.args.args)
            self.visit(node.args.vararg)
            self.visit(node.args.kwonlyargs)
            self.visit(node.args.kwarg)
            discard_set = self.rval_symbol_set
        # throw away anything appearing in lambda body that isn't bound
        self.rval_symbol_set -= discard_set

    def visit_With(self, node):
        # skip body
        self.visit(node.items)

    def visit_withitem(self, node):
        with self.gather_lvals_context():
            self.visit(node.optional_vars)
        with self.gather_rvals_context():
            self.visit(node.context_expr)

    def visit_arg(self, node):
        self.to_add_set.add(node.arg)


def get_statement_lval_and_rval_symbols(node: ast.AST):
    return GetStatementLvalRvalSymbols()(node)



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

from IPython import get_ipython

from .analysis import AttributeSymbolChain, CallPoint
from .data_cell import ClassDataCell, DataCell, FunctionDataCell

if TYPE_CHECKING:
    from typing import Dict, Optional, Set, Tuple, Union


class Scope(object):
    GLOBAL_SCOPE_NAME = '<module>'

    def __init__(
            self, scope_name: str = GLOBAL_SCOPE_NAME,
            parent_scope: 'Optional[Scope]' = None,
            is_namespace_scope=False
    ):
        self.scope_name = scope_name
        self.parent_scope = parent_scope  # None iff this is the global scope
        self.is_namespace_scope = is_namespace_scope
        self.data_cell_by_name: Dict[str, DataCell] = {}

    def __hash__(self):
        return hash(self.full_path)

    def __str__(self):
        return str(self.full_path)

    def clone(self):
        cloned = Scope()
        cloned.__dict__ = dict(self.__dict__)
        # we don't want copies of the data cells but aliases instead,
        # but we still want separate dictionaries for newly created DataCells
        cloned.data_cell_by_name = dict(self.data_cell_by_name)
        return cloned

    @property
    def non_namespace_parent_scope(self):
        # a scope nested inside of a namespace scope does not have access
        # to unqualified members of the namespace scope
        if self.is_global:
            return None
        if self.parent_scope.is_namespace_scope:
            return self.parent_scope.non_namespace_parent_scope
        return self.parent_scope

    def make_child_scope(self, scope_name, is_namespace_scope=False):
        return self.__class__(scope_name, parent_scope=self, is_namespace_scope=is_namespace_scope)

    def lookup_data_cell_by_name(self, name):
        ret = self.data_cell_by_name.get(name, None)
        if ret is None and self.non_namespace_parent_scope is not None:
            ret = self.non_namespace_parent_scope.lookup_data_cell_by_name(name)
        return ret

    def gen_data_cells_for_attr_symbol_chain(self, chain: AttributeSymbolChain, namespaces: 'Dict[int, Scope]'):
        cur_scope = self
        name_to_obj = get_ipython().ns_table['user_global']
        for name in chain.symbols:
            if isinstance(name, CallPoint):
                break
            dc = cur_scope.data_cell_by_name.get(name, None)
            if dc is None:
                break
            yield dc
            obj = name_to_obj.get(name, None)
            if obj is None:
                break
            cur_scope = namespaces.get(id(obj), None)
            if cur_scope is None:
                break
            name_to_obj = obj.__dict__

    def _upsert_and_mark_children_if_different_data_cell_type(
            self, dc: 'Union[ClassDataCell, FunctionDataCell]', name: str, deps: 'Set[DataCell]'
    ):
        if self.is_globally_accessible and name in self.data_cell_by_name:
            old = self.data_cell_by_name[name]
            # don't mark children as having stale dep unless old dep was of same type
            old.update_deps(set(), add=False, mark_children=isinstance(old, type(dc)))
        dc.update_deps(deps, add=False)
        self.data_cell_by_name[name] = dc
        return dc

    def _upsert_function_data_cell_for_name(self, name: str, deps: 'Set[DataCell]'):
        dc = FunctionDataCell(self.make_child_scope(name), name)
        return self._upsert_and_mark_children_if_different_data_cell_type(dc, name, deps)

    def _upsert_class_data_cell_for_name(self, name: str, deps: 'Set[DataCell]', class_scope: 'Scope'):
        dc = ClassDataCell(class_scope, name)
        return self._upsert_and_mark_children_if_different_data_cell_type(dc, name, deps)

    def upsert_data_cell_for_name(
            self,
            name: str,
            deps: 'Set[DataCell]',
            add=False,
            is_function_def=False,
            class_scope: 'Optional[Scope]' = None,
    ):
        assert not (class_scope is not None and is_function_def)
        if is_function_def:
            assert not add
            return self._upsert_function_data_cell_for_name(name, deps)
        if class_scope is not None:
            assert not add
            return self._upsert_class_data_cell_for_name(name, deps, class_scope)
        if self.is_globally_accessible and name in self.data_cell_by_name:
            # TODO: handle case where new dc is of different type
            dc = self.data_cell_by_name[name]
            dc.update_deps(deps, add=add)
            # TODO: garbage collect old names
            return dc
        dc = DataCell(name, deps)
        self.data_cell_by_name[name] = dc
        for dep in deps:
            dep.children.add(dc)
        return dc

    @property
    def is_global(self):
        return self.parent_scope is None

    @property
    def is_globally_accessible(self):
        return self.is_global or (self.is_namespace_scope and self.parent_scope.is_globally_accessible)

    @property
    def global_scope(self):
        if self.is_global:
            return self
        return self.parent_scope.global_scope

    @property
    def full_path(self) -> 'Tuple[str, ...]':
        path = (self.scope_name,)
        if self.is_global:
            return path
        else:
            return self.parent_scope.full_path + path



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

import networkx as nx

if TYPE_CHECKING:
    from typing import List
    from .safety import DependencySafety


USAGE = """Options:

show_graph: 
    - This will print out the dependency graph of global variables. Stale nodes are labeled red.
      Notice that user might need to call this twice to have it to work.

show_dependency <variable_name> <variable_name2> ...: 
    - This will print out the dependencies for given global variables.
      Multiple variables should be separated with spaces.

show_stale: 
    - This will print out all the global variables that are stale. 

set_disable_level <integer>:
    - level 0: Warning,    Stop Code,   Record new dependencies, (Full functionality)
    - level 1: Warning,    Run code,    Record new dependencies, (Don't stop at the warning)
    - level 2: No Warning, Run code,    Record new dependencies, (Don't show warnings)
    - level 3: No warning, Run cocde,   No new dependencies,     (Original Kernel)

remove_dependency <parent_name> <child_name>:
    - This will remove the dependency between parent variable and the child variable.

add_dependency <parent_name> <child_name>:
    - This will add the dependency between parent variable and the child variable.

turn_off_warnings_for  <variable_name> <variable_name2> ...: 
    - This will turn off the warnings for given global variables. These variables will not be
      considered as stale anymore. Multiple variables should be separated with spaces.


turn_off_warnings_for  <variable_name> <variable_name2> ...: 
    - This will turn the warnings back on for given global variables. These variables could have
      stale dependencies now. Multiple variables should be separated with spaces."""


def show_graph(safety: 'DependencySafety'):
    graph = nx.DiGraph()
    for name in safety.global_scope.data_cell_by_name:
        graph.add_node(name)
    for node in safety.global_scope.data_cell_by_name.values():
        name = node.name
        for child_node in node.children:
            graph.add_edge(name, child_node.name)
    nx.draw_networkx(
        graph,
        node_color=["#ff0000" if safety.global_scope.lookup_data_cell_by_name(name).is_stale() else "#cccccc" for name in graph.nodes()],
        arrowstyle='->',
        arrowsize=30,
        node_size=1000,
        pos=nx.drawing.layout.planar_layout(graph)
    )


def show_deps(safety: 'DependencySafety', line: 'List[str]'):
    if len(line) == 1:
        print("Usage: %safety show_dependency <variable_name> <variable_name2> ...")
        return
    for data_cell_name in line[1:]:
        data_cell = safety.global_scope.lookup_data_cell_by_name(data_cell_name)
        if data_cell:
            print("DataCell {} is dependent on {}".format(data_cell_name, [str(n) for n in data_cell.parents] if data_cell.parents else "Nothing"))
        else:
            print("Cannot find DataCell", data_cell_name)


def show_stale(safety: 'DependencySafety'):
    stale_set = set()
    for data_cell in safety.global_scope.data_cell_by_name.values():
        if data_cell.is_stale():
            stale_set.add(data_cell)
    if not stale_set:
        print("No DataCell has stale dependency for now!")
    elif len(stale_set) == 1:
        print("The only DataCell with stale dependencies is:", str(stale_set.pop()))
    else:
        print("DataCells with stale dependencies are:", [str(n) for n in stale_set])


def set_disable_level(safety: 'DependencySafety', line: 'List[str]'):
    if len(line) != 2 or line[1] not in ['0', '1', '2', '3']:
        print("""Usage: %safety set_disable_level <integer>
-level 0: Warning,    Stop Code,   Record new dependencies, (Full functionality)
-level 1: Warning,    Run code,    Record new dependencies, (Don't stop at the warning)
-level 2: No Warning, Run code,    Record new dependencies, (Don't show warnings)
-level 3: No warning, Run cocde,   No new dependencies,     (Original Kernel)""")
        return
    safety._disable_level = int(line[1])


def configure_trace_messages(safety: 'DependencySafety', line: 'List[str]'):
    if len(line) != 2:
        print("Usage: %safety trace_messages [enabled|disabled] ...")
        return
    safety.trace_messages_enabled = (line[1].lower().startswith("enable"))


def remove_dep(safety: 'DependencySafety', line: 'List[str]'):
    if len(line) != 3:
        print("Usage: %safety remove_dependency <parent_name> <child_name>")
        return
    parent_data_cell = safety.global_scope.lookup_data_cell_by_name(line[1])
    if not parent_data_cell:
        print("Cannot find DataCell", line[1])
        return
    child_data_cell = safety.global_scope.lookup_data_cell_by_name(line[2])
    if not child_data_cell:
        print("Cannot find DataCell", line[2])
        return
    if child_data_cell not in parent_data_cell.children or parent_data_cell not in child_data_cell.parents:
        print("Two DataCells do not have a dependency relation")
        return
    parent_data_cell.children.remove(child_data_cell)
    child_data_cell.parents.remove(parent_data_cell)


def add_dep(safety: 'DependencySafety', line: 'List[str]'):
    if len(line) != 3:
        print("Usage: %safety add_dependency <parent_name> <child_name>")
        return
    parent_data_cell = safety.global_scope.lookup_data_cell_by_name(line[1])
    if not parent_data_cell:
        print("Cannot find DataCell", line[1])
        return
    child_data_cell = safety.global_scope.lookup_data_cell_by_name(line[2])
    if not child_data_cell:
        print("Cannot find DataCell", line[2])
        return
    if child_data_cell in parent_data_cell.children and parent_data_cell in child_data_cell.parents:
        print("Two DataCells already have a dependency relation")
        return
    parent_data_cell.children.add(child_data_cell)
    child_data_cell.parents.add(parent_data_cell)


def turn_off_warnings_for(safety: 'DependencySafety', line: 'List[str]'):
    if len(line) <= 1:
        print("Usage: %safety turn_off_warnings_for <variable_name> <variable_name2> ...")
        return
    for data_cell_name in line[1:]:
        data_cell = safety.global_scope.lookup_data_cell_by_name(data_cell_name)
        if data_cell:
            data_cell.no_warning = True
            print("Warnings are turned off for", data_cell_name)
        else:
            print("Cannot find DataCell", data_cell_name)


def turn_on_warnings_for(safety: 'DependencySafety', line: 'List[str]'):
    if len(line) <= 1:
        print("Usage: %safety turn_on_warnings_for <variable_name> <variable_name2> ...")
        return
    for data_cell_name in line[1:]:
        data_cell = safety.global_scope.lookup_data_cell_by_name(data_cell_name)
        if data_cell:
            data_cell.no_warning = False
            print("Warnings are turned on for", data_cell_name)
        else:
            print("Cannot find DataCell", data_cell_name)



# -*- coding: utf-8 -*-

from nbsafety.analysis.stmt_edges import get_statement_lval_and_rval_symbols
from nbsafety.analysis.lineno_stmt_map import compute_lineno_to_stmt_mapping


def test_classes():
    code = """
class Foo(object):
    pass
    
class Bar(Foo):
    pass
    
class Baz(Foo, Bar):
    pass
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'Foo'}
    assert rvals == {'object'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[4])
    assert lvals == {'Bar'}
    assert rvals == {'Foo'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[7])
    assert lvals == {'Baz'}
    assert rvals == {'Foo', 'Bar'}


def test_for_loop():
    code = """
for i in range(10):
    a = i
    b = a + i
    lst = [a, b]
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'i'}
    assert rvals == {'range'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[2])
    assert lvals == {'a'}
    assert rvals == {'i'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[3])
    assert lvals == {'b'}
    assert rvals == {'a', 'i'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[4])
    assert lvals == {'lst'}
    assert rvals == {'a', 'b'}


def test_context_manager():
    code = """
fname = 'file.txt'
with open(fname) as f:
    contents = f.read()
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[2])
    assert lvals == {'f'}
    assert rvals == {'fname', 'open'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[3])
    assert lvals == {'contents'}
    assert rvals == {'f'}  # attributes should be skipped



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import os
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import check_call

pkg_name = 'nbsafety'


def make_post_install_hook(install_or_develop):
    class PostInstallHook(install_or_develop):
        """Post-installation for installation mode."""
        def run(self):
            install_or_develop.run(self)
            python_command_suffix = ''
            if platform.system().lower().startswith('win'):
                python_command_suffix = '.exe'
            check_call(f'python{python_command_suffix} -m {pkg_name}.install'.split())
    return PostInstallHook


def read_file(fname):
    with open(fname, 'r') as f:
        return f.read()


history = read_file('HISTORY.rst')
requirements = read_file('requirements.txt').strip().split()
exec(read_file(os.path.join(pkg_name, 'version.py')))
setup(
    name=pkg_name,
    version=__version__,  # noqa
    author='Stephen Macke, Ray Gong',
    author_email='stephen.macke@gmail.com',
    description='Fearless interactivity for Jupyter notebooks.',
    long_description=read_file('README.md'),
    url='https://github.com/runtime-jupyter-safety/nbsafety',  # maybe rename to nbsafety
    packages=find_packages(exclude=['docs', 'scratchspace', 'notebooks', 'img', 'test']),
    include_package_data=True,
    install_requires=requirements,
    license='BSD-3-Clause',
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    cmdclass={
        'develop': make_post_install_hook(develop),
        'install': make_post_install_hook(install),
    },
)

# python setup.py sdist bdist_wheel --universal
# twine upload dist/*



# -*- coding: utf-8 -*-
import ast
import builtins
from contextlib import contextmanager
import logging
from typing import TYPE_CHECKING

from ..data_cell import DataCell

if TYPE_CHECKING:
    from typing import Dict, List, Set, Tuple, Union
    from ..scope import Scope

logger = logging.getLogger(__name__)


class AttributeTracingManager(object):
    def __init__(self, namespaces: 'Dict[int, Scope]', active_scope: 'Scope'):
        self.namespaces = namespaces
        self.original_active_scope = active_scope
        self.active_scope = active_scope
        self.start_tracer_name = '_ATTR_TRACER_START'
        self.end_tracer_name = '_ATTR_TRACER_END'
        setattr(builtins, self.start_tracer_name, self.attribute_tracer)
        setattr(builtins, self.end_tracer_name, self.expr_tracer)
        self.ast_transformer = AttributeTracingNodeTransformer(self.start_tracer_name, self.end_tracer_name)
        self.loaded_data_cells: Set[DataCell] = set()
        self.stored_scope_qualified_names: Set[Tuple[Scope, str]] = set()
        self.aug_stored_scope_qualified_names: Set[Tuple[Scope, str]] = set()
        self.stack: List[Tuple[Set[Tuple[Scope, str]], Set[Tuple[Scope, str]], Scope, Scope]] = []

    def __del__(self):
        if hasattr(builtins, self.start_tracer_name):
            delattr(builtins, self.start_tracer_name)
        if hasattr(builtins, self.end_tracer_name):
            delattr(builtins, self.end_tracer_name)

    def push_stack(self, new_scope: 'Scope'):
        self.stack.append((
            self.stored_scope_qualified_names,
            self.aug_stored_scope_qualified_names,
            self.active_scope,
            self.original_active_scope,
        ))
        self.stored_scope_qualified_names = set()
        self.aug_stored_scope_qualified_names = set()
        self.original_active_scope = new_scope
        self.active_scope = new_scope

    def pop_stack(self):
        (
            self.stored_scope_qualified_names,
            self.aug_stored_scope_qualified_names,
            self.active_scope,
            self.original_active_scope,
        ) = self.stack.pop()

    @staticmethod
    def debug_attribute_tracer(obj, attr, ctx):
        logger.debug('%s attr %s of obj %s', ctx, attr, obj)
        return obj

    def attribute_tracer(self, obj, attr, ctx, override_active_scope):
        obj_id = id(obj)
        scope = self.namespaces.get(obj_id, None)
        # print('%s attr %s of obj %s' % (ctx, attr, obj))
        if scope is None:
            class_scope = self.namespaces.get(id(obj.__class__), None)
            if class_scope is not None:
                # print('found class scope %s containing %s' % (class_scope, class_scope.data_cell_by_name.keys()))
                scope = class_scope.clone()
                self.namespaces[obj_id] = scope
            # else:
            #     print('no scope for class', obj.__class__)
        # print('new active scope', scope)
        if override_active_scope:
            self.active_scope = scope
        if scope is None:
            return obj
        if ctx == 'Load':
            data_cell = scope.data_cell_by_name.get(attr, None)
            if data_cell is None:
                data_cell = DataCell(attr)
                scope.data_cell_by_name[attr] = data_cell
            self.loaded_data_cells.add(data_cell)
        if ctx == 'Store':
            self.stored_scope_qualified_names.add((scope, attr))
        if ctx == 'AugStore':
            self.aug_stored_scope_qualified_names.add((scope, attr))
        return obj

    def expr_tracer(self, obj):
        # print('reset active scope to', self.original_active_scope)
        self.active_scope = self.original_active_scope
        return obj

    def reset(self):
        self.loaded_data_cells = set()
        self.stored_scope_qualified_names = set()
        self.aug_stored_scope_qualified_names = set()
        self.active_scope = self.original_active_scope


class AttributeTracingNodeTransformer(ast.NodeTransformer):
    def __init__(self, start_tracer: str, end_tracer: str):
        self.start_tracer = start_tracer
        self.end_tracer = end_tracer
        self.inside_attr_load_chain = False

    @contextmanager
    def attribute_load_context(self, override=True):
        old = self.inside_attr_load_chain
        self.inside_attr_load_chain = override or old
        yield
        self.inside_attr_load_chain = old

    def visit_Attribute(self, node: 'ast.Attribute'):
        override_active_scope = isinstance(node.ctx, ast.Load)
        override_active_scope_arg = ast.Constant(override_active_scope)
        ast.copy_location(override_active_scope_arg, node)
        with self.attribute_load_context(override_active_scope):
            replacement_value = ast.Call(
                func=ast.Name(self.start_tracer, ctx=ast.Load()),
                args=[
                    self.visit(node.value), ast.Str(node.attr),
                    ast.Str(node.ctx.__class__.__name__), override_active_scope_arg
                ],
                keywords=[]
            )
        ast.copy_location(replacement_value, node.value)
        node.value = replacement_value
        new_node: Union[ast.Attribute, ast.Call] = node
        if not self.inside_attr_load_chain and override_active_scope:
            new_node = ast.Call(
                func=ast.Name(self.end_tracer, ctx=ast.Load()),
                args=[node],
                keywords=[]
            )
        return new_node

    def visit_Call(self, node: ast.Call):
        if not isinstance(node.func, ast.Attribute):
            return node
        assert isinstance(node.func.ctx, ast.Load)
        with self.attribute_load_context():
            node.func = self.visit_Attribute(node.func)
        if self.inside_attr_load_chain:
            return node
        replacement_node = ast.Call(
            func=ast.Name(self.end_tracer, ctx=ast.Load()),
            args=[node],
            keywords=[]
        )
        ast.copy_location(replacement_node, node)
        return replacement_node
